'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; }; /**
                                                                                                                                                                                                                                                                   *  Copyright (c) Facebook, Inc.
                                                                                                                                                                                                                                                                   *  All rights reserved.
                                                                                                                                                                                                                                                                   *
                                                                                                                                                                                                                                                                   *  This source code is licensed under the license found in the
                                                                                                                                                                                                                                                                   *  LICENSE file in the root directory of this source tree.
                                                                                                                                                                                                                                                                   *
                                                                                                                                                                                                                                                                   *  
                                                                                                                                                                                                                                                                   */

exports.processIPCNotificationMessage = processIPCNotificationMessage;
exports.processIPCRequestMessage = processIPCRequestMessage;
exports.processStreamMessage = processStreamMessage;

var _graphqlLanguageServiceConfig = require('graphql-language-service-config');

var _graphqlLanguageServiceInterface = require('graphql-language-service-interface');

var _graphqlLanguageServiceUtils = require('graphql-language-service-utils');

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _GraphQLCache = require('./GraphQLCache');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Response message error codes
var ERROR_CODES = {
  PARSE_ERROR: -32700,
  INVALID_REQUEST: -32600,
  METHOD_NOT_FOUND: -32601,
  INVALID_PARAMS: -32602,
  INTERNAL_ERROR: -32603,
  SERVER_ERROR_START: -32099,
  SERVER_ERROR_END: -32000,
  SERVER_NOT_INITIALIZED: -32002,
  UNKNOWN_ERROR_CODE: -32001
};

var REQUEST_IDS_IN_PROGRESS = [];

var graphQLCache = void 0;
var languageService = void 0;
var textDocumentCache = new Map();

function processIPCNotificationMessage(message, writer) {
  var method, response, textDocument, uri, text, cachedDocument, diagnostics, contentChanges;
  return regeneratorRuntime.async(function processIPCNotificationMessage$(_context) {
    while (1) {
      switch (_context.prev = _context.next) {
        case 0:
          method = message.method;
          response = void 0;
          textDocument = void 0;
          _context.t0 = method;
          _context.next = _context.t0 === 'textDocument/didOpen' ? 6 : _context.t0 === 'textDocument/didSave' ? 6 : _context.t0 === 'textDocument/didChange' ? 18 : _context.t0 === 'textDocument/didClose' ? 24 : _context.t0 === 'exit' ? 29 : 31;
          break;

        case 6:
          if (!(!message.params || !message.params.textDocument)) {
            _context.next = 8;
            break;
          }

          return _context.abrupt('return');

        case 8:
          textDocument = message.params.textDocument;
          uri = textDocument.uri;
          text = textDocument.text;

          // Create/modify the cached entry if text is provided.
          // Otherwise, try searching the cache to perform diagnostics.

          if (text) {
            invalidateCache(textDocument, uri, { text: text });
          } else {
            if (textDocumentCache.has(uri)) {
              cachedDocument = textDocumentCache.get(uri);

              if (cachedDocument) {
                text = cachedDocument.content.text;
              }
            }
          }

          _context.next = 14;
          return regeneratorRuntime.awrap(provideDiagnosticsMessage(text, uri));

        case 14:
          diagnostics = _context.sent;

          response = convertToRpcMessage({
            method: 'textDocument/publishDiagnostics',
            params: { uri: uri, diagnostics: diagnostics }
          });
          sendMessageIPC(response, writer);
          return _context.abrupt('break', 31);

        case 18:
          if (!(!message.params || !message.params.textDocument || !message.params.contentChanges)) {
            _context.next = 20;
            break;
          }

          return _context.abrupt('return');

        case 20:

          textDocument = message.params.textDocument;
          contentChanges = message.params.contentChanges;

          // As `contentChanges` is an array and we just want the
          // latest update to the text, grab the last entry from the array.

          invalidateCache(textDocument, textDocument.uri || message.params.uri, contentChanges[contentChanges.length - 1]);
          return _context.abrupt('break', 31);

        case 24:
          if (!(!message.params || !message.params.textDocument)) {
            _context.next = 26;
            break;
          }

          return _context.abrupt('return');

        case 26:
          textDocument = message.params.textDocument;

          if (textDocumentCache.has(textDocument.uri)) {
            textDocumentCache.delete(textDocument.uri);
          }
          return _context.abrupt('break', 31);

        case 29:
          process.exit(0);
          return _context.abrupt('break', 31);

        case 31:
        case 'end':
          return _context.stop();
      }
    }
  }, null, this);
}

function processIPCRequestMessage(message, configDir, writer) {
  var method, response, textDocument, cachedDocument, query, result, serverCapabilities, position, pos, formatted, requestIDToCancel, index;
  return regeneratorRuntime.async(function processIPCRequestMessage$(_context2) {
    while (1) {
      switch (_context2.prev = _context2.next) {
        case 0:
          method = message.method;
          response = void 0;
          textDocument = void 0;
          cachedDocument = void 0;
          query = void 0;
          result = void 0;
          _context2.t0 = method;
          _context2.next = _context2.t0 === 'initialize' ? 9 : _context2.t0 === 'textDocument/completion' ? 17 : _context2.t0 === 'textDocument/definition' ? 30 : _context2.t0 === '$/cancelRequest' ? 44 : _context2.t0 === 'shutdown' ? 50 : 51;
          break;

        case 9:
          if (!(!message.params || !message.params.rootPath)) {
            _context2.next = 11;
            break;
          }

          return _context2.abrupt('return');

        case 11:
          _context2.next = 13;
          return regeneratorRuntime.awrap(initialize(configDir ? configDir.trim() : message.params.rootPath));

        case 13:
          serverCapabilities = _context2.sent;


          if (serverCapabilities === null) {
            response = convertToRpcMessage({
              id: '-1',
              error: {
                code: ERROR_CODES.SERVER_NOT_INITIALIZED,
                message: '.graphqlrc not found'
              }
            });
          } else {
            response = convertToRpcMessage({
              id: message.id,
              result: serverCapabilities
            });
          }
          sendMessageIPC(response, writer);
          return _context2.abrupt('break', 51);

        case 17:
          if (!(!message.params || !message.params.textDocument || !message.params.position)) {
            _context2.next = 19;
            break;
          }

          return _context2.abrupt('return');

        case 19:
          textDocument = message.params.textDocument;
          position = message.params.position;


          cachedDocument = getCachedDocument(textDocument.uri);

          if (cachedDocument) {
            _context2.next = 24;
            break;
          }

          return _context2.abrupt('return');

        case 24:

          query = cachedDocument.content.text;
          _context2.next = 27;
          return regeneratorRuntime.awrap(languageService.getAutocompleteSuggestions(query, position, textDocument.uri));

        case 27:
          result = _context2.sent;

          sendMessageIPC(convertToRpcMessage({
            id: message.id,
            result: { items: result, isCompleted: true }
          }), writer);
          return _context2.abrupt('break', 51);

        case 30:
          if (!(!message.params || !message.params.textDocument || !message.params.position)) {
            _context2.next = 32;
            break;
          }

          return _context2.abrupt('return');

        case 32:
          textDocument = message.params.textDocument;
          pos = message.params.position;


          cachedDocument = getCachedDocument(textDocument.uri);

          if (cachedDocument) {
            _context2.next = 37;
            break;
          }

          return _context2.abrupt('return');

        case 37:

          query = cachedDocument.content.text;
          _context2.next = 40;
          return regeneratorRuntime.awrap(languageService.getDefinition(query, pos, textDocument.uri));

        case 40:
          result = _context2.sent;
          formatted = result ? result.definitions.map(function (res) {
            return {
              uri: _path2.default.join('file://', res.path),
              range: res.range
            };
          }) : [];

          sendMessageIPC(convertToRpcMessage({
            id: message.id,
            result: formatted
          }), writer);
          return _context2.abrupt('break', 51);

        case 44:
          if (!(!message.params || !message.params.id)) {
            _context2.next = 46;
            break;
          }

          return _context2.abrupt('return');

        case 46:
          requestIDToCancel = message.params.id;
          index = REQUEST_IDS_IN_PROGRESS.indexOf(requestIDToCancel);

          if (index !== -1) {
            REQUEST_IDS_IN_PROGRESS.splice(index, 1);
            // A cancelled request still needs to send an empty response back
            sendMessageIPC({ id: requestIDToCancel }, writer);
          }
          return _context2.abrupt('break', 51);

        case 50:
          return _context2.abrupt('break', 51);

        case 51:
        case 'end':
          return _context2.stop();
      }
    }
  }, null, this);
}

function processStreamMessage(message, configDir) {
  var graphQLConfigDir, json, id, method, result, responseMsg, _json$args, query, filePath, position, formatted;

  return regeneratorRuntime.async(function processStreamMessage$(_context3) {
    while (1) {
      switch (_context3.prev = _context3.next) {
        case 0:
          if (!(message.length === 0)) {
            _context3.next = 2;
            break;
          }

          return _context3.abrupt('return');

        case 2:
          if (graphQLCache) {
            _context3.next = 10;
            break;
          }

          graphQLConfigDir = (0, _graphqlLanguageServiceConfig.findGraphQLConfigDir)(configDir ? configDir.trim() : process.cwd());

          if (graphQLConfigDir) {
            _context3.next = 7;
            break;
          }

          process.stdout.write(JSON.stringify(convertToRpcMessage({
            id: '-1',
            error: {
              code: ERROR_CODES.SERVER_NOT_INITIALIZED,
              message: '.graphqlrc not found'
            }
          })));
          return _context3.abrupt('return');

        case 7:
          _context3.next = 9;
          return regeneratorRuntime.awrap((0, _GraphQLCache.getGraphQLCache)(graphQLConfigDir));

        case 9:
          graphQLCache = _context3.sent;

        case 10:
          if (!languageService) {
            languageService = new _graphqlLanguageServiceInterface.GraphQLLanguageService(graphQLCache);
          }

          json = void 0;
          _context3.prev = 12;

          json = JSON.parse(message);
          _context3.next = 20;
          break;

        case 16:
          _context3.prev = 16;
          _context3.t0 = _context3['catch'](12);

          process.stdout.write(JSON.stringify(convertToRpcMessage({
            id: '-1',
            error: {
              code: ERROR_CODES.PARSE_ERROR,
              message: 'Request contains incorrect JSON format'
            }
          })));
          return _context3.abrupt('return');

        case 20:
          id = json.id;
          method = json.method;
          result = null;
          responseMsg = null;
          _json$args = json.args, query = _json$args.query, filePath = _json$args.filePath, position = _json$args.position;
          _context3.t1 = method;
          _context3.next = _context3.t1 === 'disconnect' ? 28 : _context3.t1 === 'getDiagnostics' ? 30 : _context3.t1 === 'getDefinition' ? 36 : _context3.t1 === 'getAutocompleteSuggestions' ? 42 : _context3.t1 === 'getOutline' ? 49 : 53;
          break;

        case 28:
          process.exit(0);
          return _context3.abrupt('break', 54);

        case 30:
          _context3.next = 32;
          return regeneratorRuntime.awrap(provideDiagnosticsMessage(query, filePath));

        case 32:
          result = _context3.sent;

          responseMsg = convertToRpcMessage({
            type: 'response',
            id: id,
            result: result
          });
          process.stdout.write(JSON.stringify(responseMsg) + '\n');
          return _context3.abrupt('break', 54);

        case 36:
          _context3.next = 38;
          return regeneratorRuntime.awrap(languageService.getDefinition(query, position, filePath));

        case 38:
          result = _context3.sent;

          responseMsg = convertToRpcMessage({
            type: 'response',
            id: id,
            result: result
          });
          process.stdout.write(JSON.stringify(responseMsg) + '\n');
          return _context3.abrupt('break', 54);

        case 42:
          _context3.next = 44;
          return regeneratorRuntime.awrap(languageService.getAutocompleteSuggestions(query, position, filePath));

        case 44:
          result = _context3.sent;
          formatted = result.map(function (res) {
            return {
              text: res.label,
              typeName: res.detail ? String(res.detail) : null,
              description: res.documentation || null
            };
          });

          responseMsg = convertToRpcMessage({
            type: 'response',
            id: id,
            formatted: formatted
          });
          process.stdout.write(JSON.stringify(responseMsg) + '\n');
          return _context3.abrupt('break', 54);

        case 49:
          result = (0, _graphqlLanguageServiceInterface.getOutline)(query);
          responseMsg = convertToRpcMessage({
            type: 'response',
            id: id,
            result: result
          });
          process.stdout.write(JSON.stringify(responseMsg) + '\n');
          return _context3.abrupt('break', 54);

        case 53:
          return _context3.abrupt('break', 54);

        case 54:
        case 'end':
          return _context3.stop();
      }
    }
  }, null, this, [[12, 16]]);
}

/**
 * Helper functions to perform requested services from client/server.
 */

function initialize(rootPath) {
  var serverCapabilities, configDir;
  return regeneratorRuntime.async(function initialize$(_context4) {
    while (1) {
      switch (_context4.prev = _context4.next) {
        case 0:
          serverCapabilities = {
            capabilities: {
              completionProvider: { resolveProvider: true },
              definitionProvider: true,
              textDocumentSync: 1
            }
          };
          configDir = (0, _graphqlLanguageServiceConfig.findGraphQLConfigDir)(rootPath);

          if (configDir) {
            _context4.next = 4;
            break;
          }

          return _context4.abrupt('return', null);

        case 4:
          _context4.next = 6;
          return regeneratorRuntime.awrap((0, _GraphQLCache.getGraphQLCache)(configDir));

        case 6:
          graphQLCache = _context4.sent;

          languageService = new _graphqlLanguageServiceInterface.GraphQLLanguageService(graphQLCache);

          return _context4.abrupt('return', serverCapabilities);

        case 9:
        case 'end':
          return _context4.stop();
      }
    }
  }, null, this);
}

function provideDiagnosticsMessage(query, uri) {
  var results, queryLines, totalLines, lastLineLength, lastCharacterPosition;
  return regeneratorRuntime.async(function provideDiagnosticsMessage$(_context5) {
    while (1) {
      switch (_context5.prev = _context5.next) {
        case 0:
          _context5.next = 2;
          return regeneratorRuntime.awrap(languageService.getDiagnostics(query, uri));

        case 2:
          results = _context5.sent;

          if (results && results.length > 0) {
            queryLines = query.split('\n');
            totalLines = queryLines.length;
            lastLineLength = queryLines[totalLines - 1].length;
            lastCharacterPosition = new _graphqlLanguageServiceUtils.Position(totalLines, lastLineLength);

            results = results.filter(function (diagnostic) {
              return diagnostic.range.end.lessThanOrEqualTo(lastCharacterPosition);
            });
          }

          return _context5.abrupt('return', results);

        case 5:
        case 'end':
          return _context5.stop();
      }
    }
  }, null, this);
}

function sendMessageIPC(message, writer) {
  writer.write(message);
}

/**
 * Composes a language server protocol JSON message.
 */
function convertToRpcMessage(metaMessage) {
  var message = _extends({
    jsonrpc: '2.0',
    protocol: 'graphql_language_service'
  }, metaMessage);

  return message;
}

function invalidateCache(textDocument, uri, content) {
  if (!uri) {
    return;
  }

  if (textDocumentCache.has(uri)) {
    var cachedDocument = textDocumentCache.get(uri);
    if (cachedDocument && cachedDocument.version < textDocument.version) {
      // Current server capabilities specify the full sync of the contents.
      // Therefore always overwrite the entire content.
      textDocumentCache.set(uri, {
        version: textDocument.version,
        content: content
      });
    }
  } else {
    textDocumentCache.set(uri, {
      version: textDocument.version,
      content: content
    });
  }
}

function getCachedDocument(uri) {
  if (textDocumentCache.has(uri)) {
    var cachedDocument = textDocumentCache.get(uri);
    if (cachedDocument) {
      return cachedDocument;
    }
  }

  return null;
}